# copyright: Huynh Duong Phuong Vi
# www.sketchupforyou.com
module S4U
	module S4u_scale
		class Scale_reference
			def initialize
				@scale_factor= Sketchup.version.to_i > 17 ? UI.scale_factor : 1
				@scale_factor = 1.5 if @scale_factor>1.8
				@snap=20*@scale_factor
				@delta=(1.0/(10**10))
			end
			def activate
				model=Sketchup.active_model
				@ip1=Sketchup::InputPoint.new
				@ip2=Sketchup::InputPoint.new
				@ip3=Sketchup::InputPoint.new
				@ip=Sketchup::InputPoint.new
				@drawing=false
				@sels=model.selection.select{|e| e.valid? && (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance) || e.is_a?(Sketchup::Image))}
				@entities=model.active_entities.select{|e| e && e.valid? && (e.is_a?(Sketchup::Edge)|| e.is_a?(Sketchup::Face))}-model.selection.to_a
				@edges=model.selection.select{|e| e if e.valid? && e.is_a?(Sketchup::Edge)}
				@edge_faces=model.selection.select{|e| e if e.valid? && e.is_a?(Sketchup::Face)}.collect{|f| f.edges}.flatten.uniq
				@vertices_faces=model.selection.collect{|e| e.vertices if e.valid? && e.is_a?(Sketchup::Face)}.flatten.uniq
				@vertices=model.selection.collect{|e| e.vertices if (e.valid? && (e.is_a?(Sketchup::Edge) || e.is_a?(Sketchup::Face)))}
				@vertices_trans=[]
				@vertices.flatten!
				@vertices.compact!
				@vertices.uniq!
				@tab_press=0
				if @vertices && !@vertices.empty?
					@vertices_trans=@vertices.collect{|e| [e,e.position.clone] }
				end
				@sels_trans=@sels.collect{|e| [e,e.transformation] }
				self.reset(model.active_view)
				@edit_transform=Sketchup.active_model.axes.transformation
			end
			def getExtents
				bb = Sketchup.active_model.bounds
				bb.add @ip.position if @ip.valid?
				bb.add(@p1) if @p1
				bb.add(@p2) if @p2
				return bb
			end
			def deactivate(view)
			end
			def setstatus(view=nil)
				text = Strings["Scale Reference"]
				text += " - " +  [Strings["No Uniform"] ,Strings["Uniform 2D"],Strings["Uniform 3D"]][@tab_press]
				text +=  "(" + Strings["TAB->Toggle"] + ")" + ": "
				if @state==0
					text+=Strings["Select Origin Point"]
				elsif @state==1
					text+= Strings["Select Second Point"]
				elsif @state==2
					text+= Strings["Select Reference Point(or Input Length)"]
				end
				Sketchup.status_text=text
				view.invalidate if view
			end
			def draw(view)
				model=view.model
				view.line_width=2
				view.drawing_color="red"
				if @ip && @ip.valid?
					if @ip.display?
						@ip.draw(view)
					elsif @face || @edge
						view.draw_points @ip.position, 15*@scale_factor, 1, "red"
					elsif @point
						view.draw_points @ip.position, 15*@scale_factor, 1, "red"
					end
				end
				if @state==1
					if @ip1.valid? && @ip2.valid?
						@ip1.draw(view)
						view.drawing_color ="red"
						view.draw_points @ip1.position, 15*@scale_factor, 2, "red"
						view.draw_line @ip1.position,@ip2.position
						if @ip_shift && @shift_press && @state==1
							view.line_stipple = "."
							view.drawing_color ="blue"
							view.draw_line(@ip.position, @ip2.position)
							view.line_stipple = ""
							view.drawing_color ="red"
							p2_line=@ip.position.project_to_line([@ip1.position,@ip_shift])
							view.draw_points p2_line, 15*@scale_factor, 2, "red"  if p2_line
						end
					end
				elsif @state==2
					if @p1 && @p2
						view.drawing_color ="red"
						view.draw_points @p1, 15*@scale_factor, 2, "red"
						view.draw_points @p2, 15*@scale_factor, 2, "red"
						view.draw_line @p1,@p2
					end
					if @ip.valid? && @ip1.valid? && @ip3.valid?
						@ip3.draw(view)
						view.drawing_color ="blue"
						view.draw_line @ip1.position,@ip3.position
						if @ip_shift && @shift_press
							view.line_stipple = "."
							view.drawing_color ="blue"
							view.draw_line(@ip.position, @ip3.position)
							view.line_stipple = ""
							view.drawing_color ="red"
							p2_line=@ip.position.project_to_line([@ip1.position,@ip_shift])
							view.draw_points p2_line, 15*@scale_factor, 2, "red"  if p2_line
						end
					end
				end
				#self.setstatus
			end
			def set_pointonedge(x,y,view,ip,edge)
				p1=edge.start.position.clone
				p2=edge.end.position.clone
				return nil if edge.length < (view.pixels_to_model(5*@snap, p1))
				points= [Geom.linear_combination(2.0/3, p1,1.0/3, p2),Geom.linear_combination(1.0/3, p1,2.0/3, p2)]
				ph = view.pick_helper
				ph.init(x, y)
				ph.do_pick(x, y)
				picked = points.find{|p| ph.test_point((p.transform ip.transformation),x,y,1.5*@snap) }
				if picked
					ip1=Sketchup::InputPoint.new(picked.transform ip.transformation)
					ip.copy! ip1
					view.tooltip="" + Strings["1/3"]
					return ip
				end
				return nil
			end
			def set_special_point(x,y,view,ip)
				points=[]
				ph = view.pick_helper
				ph.init(x, y)
				ph.do_pick(x, y)
				best_entity = @sels.sort_by{|e| view.screen_coords(e.bounds.center).distance view.screen_coords(ip.position) }
				best_entity=best_entity[0..7].select{|e| e.valid? && (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance) )}
				if !best_entity.empty?
					best_entity.each{|e|
						#ps= (0..7).collect{|i| e.bounds.corner(i)}
						ps= get_bounds(e)
						points+=ps
						if Sketchup.version.to_i >=14
							ps.permutation(2).to_a.each{|a|  points << Geom.linear_combination(0.5, a[0],0.5,a[1]) }
						else
							(0..7).each{|i| points+=ps.collect{|p| Geom.linear_combination(0.5, ps[i],0.5,p)} }
						end
					}
				end
				points.uniq!
				picked_bounds = points.find{|p| ph.test_point(p,x,y,1.5*@snap) }
				if picked_bounds
					ip1=Sketchup::InputPoint.new(picked_bounds)
					ip.copy! ip1
					view.tooltip=Strings["BoundingBox"]
					@point=picked_bounds
					return ip
				elsif ip.face
					cf=center_face(ip.face)
					if cf
						cf=cf.transform ip.transformation
						ph = view.pick_helper
						ph.init(x, y)
						ph.do_pick(x, y)
						picked = ph.test_point(cf,x,y,@snap)
						if picked
							ip1=Sketchup::InputPoint.new(cf)
							@face=true
							view.tooltip="" + Strings["Center Face"]
							ip.copy! ip1
						else
							@edge=ip.face.edges.find{|e| set_pointonedge(x,y,view,ip,e) }
						end
					end
				elsif ip.edge
					@edge=set_pointonedge(x,y,view,ip,ip.edge)
				end
			end
			def get_bounds(element)
				a=element.transformation.to_a
				b=nil
				if element.is_a?Sketchup::ComponentInstance
					b=element.definition.bounds
				elsif element.is_a?Sketchup::Group
					b=element.local_bounds
				else
					return nil
				end
				return (0..7).collect{|i| b.corner(i).clone.transform! element.transformation } if b
			end
			def onMouseMove(flags, x, y, view)
				@face=@edge=@point=nil
				if @state==0
					@ip.pick view,x,y
					set_special_point(x,y,view,@ip)
					view.invalidate #if @ip.valid?
				elsif @state==1
					@ip.pick view,x,y,@ip1
					set_special_point(x,y,view,@ip)
					if @shift_press && @ip_shift && @ip!=@ip1 && (@ip1.position!=@ip_shift)
						p2_line=@ip.position.project_to_line([@ip1.position,@ip_shift])
						@ip2.copy!(p2_line ? Sketchup::InputPoint.new(p2_line) : @ip)
					else
						@ip2.copy! @ip
					end
					@p2=@ip2.position
					view.invalidate #if @ip.valid?
					length = @ip1.position.distance(@ip2.position)
					Sketchup::set_status_text length.to_s, SB_VCB_VALUE
				elsif @state==2 #&& (x-@x).abs>5 && (y-@y).abs>5
					#@x,@y=x,y
					@ip.pick view,x,y,@ip1
					set_special_point(x,y,view,@ip)
					if @shift_press && @ip_shift && @ip!=@ip1 && (@ip1.position!=@ip_shift)
						p2_line=@ip.position.project_to_line([@ip1.position,@ip_shift])
						@ip3.copy!(p2_line ? Sketchup::InputPoint.new(p2_line) : @ip)
					else
						@ip3.copy! @ip
					end
					@p3=@ip3.position
					length = @ip1.position.distance(@ip3.position)
					Sketchup::set_status_text length.to_s, SB_VCB_VALUE
					# view.model.abort_operation
					# view.model.start_operation TOOL,true,false,false
					scale_elements(@sels)
					view.invalidate #if @ip.valid?
				end
				model=view.model
				entities_new=model.active_entities.select{|e| e && e.valid? && (e.is_a?(Sketchup::Edge) || e.is_a?(Sketchup::Face))}-@entities
				model.selection.add(entities_new.select{|e| e if e && e.valid? && (e.vertices-@vertices).length==0 })
			end
			def onLButtonDown(flags, x, y, view)
				if @state==0
					if @length
						@sels_trans=@sels.collect{|e| [e,e.transformation]}
						@vertices_trans=@vertices.collect{|e| [e,e.position.clone] } if !@vertices.empty?
					end
					if @ip.valid?
						@ip1.copy! @ip
						@p1=@ip1.position
						@state=1
						@length=nil
						view.lock_inference if  view.inference_locked?
					end
				elsif @state==1 && @ip2.valid? && (@ip1.position!=@ip2.position)
					#@x,@y=x,y
					@state=2
					@p3=@p2=@ip2.position
					@length=@p1.distance @p2
					view.lock_inference if  view.inference_locked?
					model=view.model
					model.start_operation TOOL,true,false,false
					@entities=model.active_entities.select{|e| e && e.valid? && (e.is_a?(Sketchup::Edge) || e.is_a?(Sketchup::Face))}-model.selection.to_a
				elsif @state==2
					clean_edge #if !@tab_press
					view.model.commit_operation
					reset view
				end
				self.setstatus(view)
			end
			def cal(input)
				if op=input[/[\+\-\/\*]/]
					begin
					if input[/[\+\-]/] && !(input[/\*[\-\+]/] || input[/\/[\-\+]/])
						index= input.rindex(/[\+\-]/)
					elsif input[/[\*\/]/]
						index= input.rindex(/[\*\/]/)
					end
					if index
						if index>0
							s=[input[0..(index-1)],input[(index+1)..-1]]
						else
							s=["0",input[(index+1)..-1]]
						end
						op=input[index]
						if op=="+"
							value= cal(s[0]).to_s.to_f + cal(s[1]).to_s.to_f
						elsif op=="-"
							value= cal(s[0]).to_s.to_f - cal(s[1]).to_s.to_f
						elsif op=="*"
							value= cal(s[0]).to_s.to_f * cal(s[1]).to_s.to_f
						elsif op=="/"
							value= cal(s[0]).to_s.to_f / cal(s[1]).to_s.to_f
						end
					end
					rescue
						value=parse_length(input)
					end
				else
					value = parse_length(input)
				end
				return parse_length(value)
			end
			def parse_length(input)
				(return 0) if !input
				input=input.to_s
				if !Sketchup.parse_length("1.1")
					input.gsub!(/\./, ',')
				else
					input.gsub!(/\,/, '.')
				end
				begin
					value=input.to_l
				rescue
					value=input.to_f
				end
				return value
			end
			def scale_elements(sels,length=nil)
				model=Sketchup.active_model
				entities=model.active_entities
				sels.each_index{|i|
					e=@sels_trans[i][0]
					et=@sels_trans[i][1]
					ta=et.to_a
					ta[11..13] = @p1.x,@p1.y,@p1.z
					ta_new=Geom::Transformation.new ta
					vec=@p2-@p1
					vec1=@p3-@p1
					vect=vec.transform ta_new.inverse
					vect1=vec1.transform ta_new.inverse
					origin=@p1.transform et.inverse
					sc=sx=sy=sz=1
					dis=@p2.distance @p1
					if dis!=0
						if @tab_press>0
							if length
								s= length/ dis
							else
								s= @p3.distance(@p1)/ dis
								s=-s if @p1.vector_to(@p2).angle_between(@p1.vector_to(@p3)) > Math::PI*0.5
							end
							if s.abs<10000
								sx=sy=sz=s
								sz=1 if @tab_press==1
							end
						elsif length
							s= length/ dis
							(sc= s) if s.abs<10000
							sx=sc if vect.x.abs > @delta
							sy=sc if vect.y.abs > @delta
							sz=sc if vect.z.abs > @delta
						else
							sx=(vect1.x/vect.x) if vect.x.abs > @delta
							sy=(vect1.y/vect.y) if vect.y.abs > @delta
							sz=(vect1.z/vect.z) if vect.z.abs > @delta
						end
						if sx.abs<10000 && sy.abs<10000 && sz.abs<10000
							zero=[sx,sy,sz].select{|x| x if x.abs >@delta}
							if zero.length>1
								t=Geom::Transformation.scaling @p1,sx,sy,sz
								origin_new=origin.transform et*t
								if origin_new!= @p1
									vt=@p1-origin_new
									trans=Geom::Transformation.translation vt
									e.transformation= trans*et*t
								else
									e.transformation = et*t
								end
							end
						end
					end
				}
				if @vertices && !@vertices.empty?
					sc=sx=sy=sz=1
					dis=@p2.distance @p1
					if dis!=0
						vec=@p2-@p1
						vec1=@p3-@p1
						ta=Geom::Transformation.new(@p1,vec)
						vect=vec.transform ta.inverse
						vect1=vec1.transform ta.inverse
						#ta_new=Geom::Transformation.new(@p1,vec1)
						if length
							s= length/dis
							sc=s if s.abs<10000
						elsif vect.z!=0
							s=(vect1.z/vect.z)
							sc=s if s.abs<10000
						end
						if ((sc.abs>0.0001) || length) && (sc.abs<10000)
							if @tab_press==1
								ts=Geom::Transformation.scaling 1,1,sc
							elsif @tab_press==2
								ts=Geom::Transformation.scaling sc,sc,sc
							else
								ts=Geom::Transformation.scaling 1,1,sc
							end
							vectors=[]
							vers=[]
							curves=[]
							vers_curves={}
							@vertices_trans.each_index{|i|
								v=@vertices_trans[i][1]
								vec = @vertices[i].position.vector_to(v)
								vnew=v.transform ta*ts*ta.inverse
								vm=v.vector_to(vnew)
								ecur=@vertices[i].edges.find{|d| d.curve && (@edges.include?(d) || @edge_faces.include?(d))  }
								if @vertices[i].curve_interior? || ecur
									curves << @vertices[i].edges.find{|d| d.curve}.curve
									vers_curves[@vertices[i]]=(@vertices_trans[i][1]+vm)
								else
									vectors << vm +vec
									vers << @vertices_trans[i][0]
								end
							}
							curves.uniq!
							curves.compact!
							curver_move=[]
							curves.each{|v|
								if @tab_press>1
									vers<< v.vertices[0]
									v1= (vers_curves[v.vertices[0]]-v.vertices[0].position)
									vectors<< (v1)
									vers<< v.vertices[-1]
									v1 = (vers_curves[v.vertices[-1]]-v.vertices[-1].position)
									vectors<< (v1)
								else
									points =v.vertices.collect{|v| vers_curves[v]}
									points.compact!
									curver_move<<[v,points] if points.length==v.vertices.length
								end
							}
							entities.transform_by_vectors(vers,vectors)
							curver_move.each{|c| c[0].move_vertices c[1]}
						end
					end
				end
				rescue => error
				p error
			end
			def clean_edge()
				model=Sketchup.active_model
				edges =@vertices.to_a.collect{|e| e.edges}.flatten.uniq #& @edges
				curves=edges.collect{|e| e.curve if e.valid? && e.curve && e.curve.is_a?(Sketchup::ArcCurve) &&
				e.curve.radius!=e.curve.vertices[1].position.distance(e.curve.center)}.compact.uniq
				curves.each{|c|
					points=[]
					points=c.vertices
					info=[]
					c.edges.each_index{|i| info<< [c.edges[i].layer,c.edges[i].material]}
					c.edges[0].explode_curve
					g=model.active_entities.add_group
					curve_new=g.entities.add_curve points
					curve_new.each_index{|i| curve_new[i].layer,curve_new[i].material=info[i][0],info[i][1]}
					if g.valid?
						ed=g.explode.select{|e| e if e && e.valid? && e.is_a?(Sketchup::Edge)}
						@edges+=ed
						@vertices+= ed.collect{|e| e.vertices}.flatten
						#model.selection.add(ed)
					end
				}
				edges_entities=model.active_entities.select{|e| e && e.valid? && e.is_a?(Sketchup::Edge)}
				edges_new= edges_entities- ((@entities + @edges) + @edge_faces)
				ed_remove=edges_new.select{|e| e if e && e.valid? && e.hidden? && (e.vertices-@vertices).length==0 &&
				(e.faces.length==0 ||(e.faces.length==2 && (e.faces[0].normal.samedirection?(e.faces[1].normal))))}
				model.active_entities.erase_entities(ed_remove.select{|e| e if e && e.valid?})
				entities_new=model.active_entities.select{|e| e && e.valid? && (e.is_a?(Sketchup::Edge)|| e.is_a?(Sketchup::Face))} - @entities
				model.selection.add(entities_new.select{|e| e if e && e.valid? && (e.vertices-@vertices).length==0 })
				rescue =>error
				p error
			end
			def get_vertices(face)
				return nil if not( face && (face.is_a?Sketchup::Face) && face.loops.length==1)
				ver=[]
				vertices=[]
				vertices =face.outer_loop.vertices*3
				(0..face.outer_loop.vertices.length-1).each { |i|
					j=face.outer_loop.vertices.length+i
					p1=vertices[j-1].position.clone
					p2=vertices[j].position.clone
					p3=vertices[j+1].position.clone
					v1=p1.vector_to p2
					v2=p2.vector_to p3
					ver.push vertices[j] if not( v1.samedirection? v2)
				}
				return ver
				rescue
				return nil
			end
			def onKeyDown(key, repeat, flags, view)
				if key==CONSTRAIN_MODIFIER_KEY
					@shift_press=true
					if view.inference_locked?
						view.lock_inference
					elsif @state==0 #&& @ip.valid?
						view.lock_inference @ip
					elsif @state==1 #&& @ip2.valid?
						@ip_shift=@ip.position
						view.lock_inference @ip2,@ip1
					elsif @state==2 #&& @ip3.valid?
						@ip_shift=@ip.position
						view.lock_inference @ip3,@ip1
					end
				elsif key==VK_UP
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p= ((@state==1) ? @p1 : @p2)
						p1=p + @edit_transform.zaxis
						ip1 = Sketchup::InputPoint.new(p1)
						ip = Sketchup::InputPoint.new(p)
						view.lock_inference ip1,ip
					end
				elsif key==VK_LEFT
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p=  ((@state==1) ? @p1 : @p2)
						p1=p + @edit_transform.yaxis
						ip1 = Sketchup::InputPoint.new(p1)
						ip = Sketchup::InputPoint.new(p)
						view.lock_inference ip1,ip
					end
				elsif key==VK_RIGHT
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p= ((@state==1) ? @p1 : @p2)
						p1=p + @edit_transform.xaxis
						ip1 = Sketchup::InputPoint.new(p1)
						ip = Sketchup::InputPoint.new(p)
						view.lock_inference ip1,ip
					end
				end
				self.setstatus(view)
			end
			def onKeyUp(key, repeat, flags, view)
				if (key==CONSTRAIN_MODIFIER_KEY)
					@shift_press=false
					view.lock_inference if view.inference_locked?
				elsif key==9
					@tab_press = @tab_press==2 ? 0 : (@tab_press+1)
					if @state==2
						model=view.model
						model.abort_operation
						model.start_operation TOOL,true,false,false
						@entities=model.active_entities.select{|e| e && e.valid? && (e.is_a?(Sketchup::Edge)|| e.is_a?(Sketchup::Face))}-model.selection.to_a
						scale_elements(@sels)
					end
				end
				self.setstatus(view)
			end
			def onUserText(text, view)
				model=view.model
				if @state==0
					if @length
						length=@length
						if text[0] =~ /[\+\-\*\/]/
							text1=text[1..-1]
							if text[0]=~ /\*/
								length= @length * text1.to_f
							elsif text[0]=~ /\//
								length= @length / text1.to_f
							elsif text[0]=~ /\+/
								length= @length + self.cal(text1)
							elsif text[0]=~ /\-/
								length= @length - self.cal(text1)
							end
						else
							length=self.cal(text)
						end
						Sketchup::set_status_text self.cal(length.to_s), SB_VCB_VALUE
						model.start_operation TOOL,true,false,false
						@entities=model.active_entities.select{|e| e && e.valid? && (e.is_a?(Sketchup::Edge)|| e.is_a?(Sketchup::Face))}-model.selection.to_a
						scale_elements(@sels,length)
						clean_edge #if !@tab_press
						model.commit_operation
					end
				elsif @state==1
					if @ip.position!=@p1
						vec=@ip.position - @p1
						length=self.cal(text)
						Sketchup::set_status_text length.to_s, SB_VCB_VALUE
						if length!=0 && vec.length>0
							vec.length=length
							@p2=@p1 + vec
							ip = Sketchup::InputPoint.new(@p2)
							@ip2.copy! ip
							@state=2
							view.lock_inference if  view.inference_locked?
						end
					end
				elsif @state==2
					@length=@p1.distance @p2 if !@length
					if text[0] =~ /[\+\-\*\/]/
						text1=text[1..-1]
						if text[0]=~ /\*/
							length= @length * text1.to_f
						elsif text[0]=~ /\//
							length= @length / text1.to_f
						elsif text[0]=~ /\+/
							length= @length + self.cal(text1)
						elsif text[0]=~ /\-/
							length= @length - self.cal(text1)
						end
					else
						length=self.cal(text)
					end
					Sketchup::set_status_text self.cal(length.to_s), SB_VCB_VALUE
					scale_elements(@sels,length)
					clean_edge #if !@tab_press
					model.commit_operation
					reset view
				end
				self.setstatus(view)
			end
			def center_face(face)
				vertices=get_vertices(face)
				return nil if !vertices
				numver=vertices.length
				cp=Geom::Point3d.new
				vertices.each{|v|
					cp.x=cp.x+v.position.x
					cp.y=cp.y+v.position.y
					cp.z=cp.z+v.position.z
				}
				cp.x=cp.x/ numver
				cp.y=cp.y/ numver
				cp.z=cp.z/ numver
				return cp
			end
			def onCancel(reason, view)
				if reason==0
					model=view.model
					model.abort_operation
					entities_new=model.active_entities.select{|e| e && e.valid? && (e.is_a?(Sketchup::Edge)|| e.is_a?(Sketchup::Face))} - @entities
					model.selection.add(entities_new.select{|e| e if e && e.valid? && (e.vertices-@vertices).length==0 })
					self.reset(view)
				end
			end
			def reset(view=nil)
				@state=0
				@face=@edge=nil
				@drawing=false
				@ip1.clear
				@ip2.clear
				@ip3.clear
				@ip.clear
				view.lock_inference if view && view.inference_locked?
				self.setstatus(view)
			end
			def resume(view)
				self.setstatus(view)
			end
		end
		def self.scale_reference
			model=Sketchup.active_model
			sels=model.selection.select{|e| e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance) || e.is_a?(Sketchup::Image) || e.is_a?(Sketchup::Edge) || e.is_a?(Sketchup::Face)}
			if  sels.empty?
				UI.messagebox Strings["Select Objects"]
				return
			end
			Sketchup.active_model.select_tool Scale_reference.new
		end
		def self.main(mode)
			self.scale_reference
		end
	end
end